from flask import current_app
from flask import url_for

from canaille.app.i18n import gettext as _
from canaille.app.mails import logo
from canaille.app.mails import send_email
from canaille.app.templating import render_template
from canaille.backends import Backend

RESET_CODE_LENGTH = 6


def send_test_mail(email):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("Test email from {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/test.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
    )
    html_body = render_template(
        "core/mails/test.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_password_reset_mail(user, mail):
    base_url = url_for("core.account.index", _external=True)
    server_name = current_app.config.get("SERVER_NAME")
    logo_cid, logo_filename, logo_raw = logo()
    subject = _("Password reset on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )

    reset_token = None
    reset_url = None
    reset_code = None
    if current_app.features.has_trusted_hosts:
        reset_token = user.generate_url_safe_token()
        Backend.instance.save(user)
        reset_url = url_for(
            "core.auth.password.reset",
            user=user,
            token=reset_token,
            _external=True,
        )
    else:
        reset_code = user.generate_sms_or_mail_otp(length=RESET_CODE_LENGTH)
        Backend.instance.save(user)

    text_body = render_template(
        "core/mails/reset.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
        server_name=server_name,
        reset_token=reset_token,
        reset_code=reset_code,
    )
    html_body = render_template(
        "core/mails/reset.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
        server_name=server_name,
        reset_token=reset_token,
        reset_code=reset_code,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=mail,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_password_initialization_mail(user, email):
    base_url = url_for("core.account.index", _external=True)
    reset_token = user.generate_url_safe_token()
    Backend.instance.save(user)
    reset_url = url_for(
        "core.auth.password.reset",
        user=user,
        token=reset_token,
        _external=True,
    )
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("Password initialization on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/firstlogin.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
    )
    html_body = render_template(
        "core/mails/firstlogin.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        reset_url=reset_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_invitation_mail(email, registration_url):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("You have been invited to create an account on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/invitation.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
    )
    html_body = render_template(
        "core/mails/invitation.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_confirmation_email(email, confirmation_url):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("Confirm your address email on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/email-confirmation.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        confirmation_url=confirmation_url,
    )
    html_body = render_template(
        "core/mails/email-confirmation.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        confirmation_url=confirmation_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_registration_mail(email, registration_url):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("Continue your registration on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/registration.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
    )
    html_body = render_template(
        "core/mails/registration.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        registration_url=registration_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_compromised_password_check_failure_mail(
    check_password_url, user_name, user_email, hashed_password, admin_email
):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("Compromised password check failure on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/compromised_password_check_failure.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        check_password_url=check_password_url,
        user_name=user_name,
        user_email=user_email,
        hashed_password=hashed_password,
    )
    html_body = render_template(
        "core/mails/compromised_password_check_failure.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        check_password_url=check_password_url,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
        user_name=user_name,
        user_email=user_email,
        hashed_password=hashed_password,
    )

    return send_email(
        subject=subject,
        recipient=admin_email,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )


def send_one_time_password_mail(mail, otp):
    base_url = url_for("core.account.index", _external=True)
    logo_cid, logo_filename, logo_raw = logo()

    subject = _("One-time passcode authentication on {website_name}").format(
        website_name=current_app.config["CANAILLE"]["NAME"]
    )
    text_body = render_template(
        "core/mails/email_otp.txt",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        otp=otp,
    )
    html_body = render_template(
        "core/mails/email_otp.html",
        site_name=current_app.config["CANAILLE"]["NAME"],
        site_url=base_url,
        otp=otp,
        logo=f"cid:{logo_cid[1:-1]}" if logo_cid else None,
        title=subject,
    )

    return send_email(
        subject=subject,
        recipient=mail,
        text=text_body,
        html=html_body,
        attachments=[(logo_cid, logo_filename, logo_raw)] if logo_filename else None,
    )



import datetime
from dataclasses import dataclass

from flask import current_app
from flask import g
from flask import session

from canaille.app import models
from canaille.core.models import User

USER_SESSION = "sessions"


@dataclass
class UserSession:
    user: User | None = None
    last_login_datetime: datetime.datetime | None = None

    @classmethod
    def deserialize(cls, payload):
        user = current_app.backend.instance.get(models.User, payload.get("user"))
        user_is_locked = (
            user and current_app.backend.has_account_lockability() and user.locked
        )
        if not user or user_is_locked:
            return None

        return UserSession(
            user=user,
            last_login_datetime=datetime.datetime.fromisoformat(
                payload["last_login_datetime"]
            )
            if payload.get("last_login_datetime")
            else None,
        )

    def serialize(self):
        return {
            "user": self.user.id,
            "last_login_datetime": self.last_login_datetime.isoformat(),
        }


def current_user_session():
    for payload in session.get(USER_SESSION, [])[::-1]:
        if obj := UserSession.deserialize(payload):
            return obj

        session[USER_SESSION].remove(payload)

    if USER_SESSION in session and not session[USER_SESSION]:
        del session[USER_SESSION]

    return None


def save_user_session():
    session[USER_SESSION][-1] = g.session.serialize()


def login_user(user):
    """Open a session for the user."""
    now = datetime.datetime.now(datetime.timezone.utc)
    obj = UserSession(user=user, last_login_datetime=now)
    g.session = obj
    try:
        previous = (
            session[USER_SESSION]
            if isinstance(session[USER_SESSION], list)
            else [session[USER_SESSION]]
        )
        session[USER_SESSION] = previous + [obj.serialize()]
    except KeyError:
        session[USER_SESSION] = [obj.serialize()]


def logout_user():
    """Close the user session."""
    try:
        session[USER_SESSION].pop()
        if not session[USER_SESSION]:
            del session[USER_SESSION]
    except (IndexError, KeyError):
        pass

    del g.session



from urllib.parse import parse_qs
from urllib.parse import urlsplit

from joserfc import jwt

from canaille.app import models


def test_oauth_implicit(testclient, user, client, backend):
    client.token_endpoint_auth_method = "none"
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="token",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()
    assert "text/html" == res.content_type

    res.form["login"] = "user"
    res = res.form.submit(status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow()

    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    client.token_endpoint_auth_method = "client_secret_basic"
    backend.save(client)


def test_oauth_implicit_auth_method_not_none(testclient, user, client, backend):
    """Even when token_endpoint_auth_method is not none, the client should be able to not authenticate if its only grant type is implicit.

    oidc-core §9 indicates that the 'none' token endpoint authentication method is used when:

         The Client does not authenticate itself at the Token Endpoint, either because it uses only the Implicit Flow (and so does not use the Token Endpoint) or because it is a Public Client with no Client Secret or other authentication mechanism.
    """
    client.token_endpoint_auth_method = "client_secret_basic"
    client.grant_types = ["implicit"]
    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="token",
            client_id=client.client_id,
            scope="profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()
    assert "text/html" == res.content_type

    res.form["login"] = "user"
    res = res.form.submit(status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow()

    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    client.token_endpoint_auth_method = "client_secret_basic"
    client.grant_types = [
        "password",
        "authorization_code",
        "implicit",
        "hybrid",
        "refresh_token",
        "client_credentials",
        "urn:ietf:params:oauth:grant-type:jwt-bearer",
    ]

    backend.save(client)


def test_oidc_implicit(testclient, server_jwk, user, client, trusted_client, backend):
    client.token_endpoint_auth_method = "none"

    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="id_token token",
            client_id=client.client_id,
            scope="openid profile",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()
    assert "text/html" == res.content_type

    res.form["login"] = "user"
    res = res.form.submit(status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow()

    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    id_token = params["id_token"][0]
    claims = jwt.decode(id_token, server_jwk)
    assert user.user_name == claims.claims["sub"]
    assert user.formatted_name == claims.claims["name"]
    assert [client.client_id, trusted_client.client_id] == claims.claims["aud"]

    client.token_endpoint_auth_method = "client_secret_basic"
    backend.save(client)


def test_oidc_implicit_with_group(
    testclient, server_jwk, user, client, foo_group, trusted_client, backend
):
    client.token_endpoint_auth_method = "none"

    backend.save(client)

    res = testclient.get(
        "/oauth/authorize",
        params=dict(
            response_type="id_token token",
            client_id=client.client_id,
            scope="openid profile groups",
            nonce="somenonce",
            redirect_uri="https://client.test/redirect1",
        ),
    ).follow()
    assert "text/html" == res.content_type

    res.form["login"] = "user"
    res = res.form.submit(status=302).follow()

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302).follow()

    assert "text/html" == res.content_type, res.json

    res = res.form.submit(name="answer", value="accept", status=302)

    assert res.location.startswith(client.redirect_uris[0])
    params = parse_qs(urlsplit(res.location).fragment)

    access_token = params["access_token"][0]
    token = backend.get(models.Token, access_token=access_token)
    assert token is not None

    id_token = params["id_token"][0]
    claims = jwt.decode(id_token, server_jwk)
    assert user.user_name == claims.claims["sub"]
    assert user.formatted_name == claims.claims["name"]
    assert [client.client_id, trusted_client.client_id] == claims.claims["aud"]
    assert ["foo"] == claims.claims["groups"]

    client.token_endpoint_auth_method = "client_secret_basic"
    backend.save(client)



import base64
import hashlib
import hmac
import json
from base64 import b64encode
from io import BytesIO
from urllib.parse import urlparse

from flask import current_app
from flask import request

DOCUMENTATION_URL = "https://canaille.readthedocs.io"


def obj_to_b64(obj):
    return base64.b64encode(json.dumps(obj).encode("utf-8")).decode("utf-8")


def b64_to_obj(string):
    return json.loads(base64.b64decode(string.encode("utf-8")).decode("utf-8"))


def build_hash(*args):
    key = current_app.config["SECRET_KEY"].encode("utf-8")
    message = obj_to_b64(str(args)).encode("utf-8")
    return hmac.new(key, message, hashlib.sha256).hexdigest()


def default_fields():
    read = set()
    write = set()
    for acl in current_app.config["CANAILLE"]["ACL"].values():
        if not acl.get("FILTER"):
            read |= set(acl.get("READ", []))
            write |= set(acl.get("WRITE", []))

    return read, write


def get_current_domain():
    if current_app.config.get("SERVER_NAME"):
        return current_app.config.get("SERVER_NAME")

    return request.host


def get_current_mail_domain():
    if current_app.config["CANAILLE"]["SMTP"]["FROM_ADDR"]:
        return current_app.config["CANAILLE"]["SMTP"]["FROM_ADDR"].split("@")[-1]

    return get_current_domain().split(":")[0]


def validate_uri(value):
    parsed = urlparse(value)
    return (parsed.scheme in ["http", "https"] or "." in parsed.scheme) and bool(
        parsed.netloc
    )


class classproperty:
    def __init__(self, f):
        self.f = f

    def __get__(self, obj, owner):
        return self.f(owner)


def get_b64encoded_qr_image(data):
    try:
        import qrcode
    except ImportError:
        return None

    qr = qrcode.QRCode(version=1, box_size=10, border=5)
    qr.add_data(data)
    qr.make(fit=True)
    img = qr.make_image(fill_color="black", back_color="white")
    buffered = BytesIO()
    img.save(buffered)
    return b64encode(buffered.getvalue()).decode("utf-8")


def mask_email(email):
    atpos = email.find("@")
    if atpos > 0:
        return email[0] + "#####" + email[atpos - 1 :]
    return None


def mask_phone(phone):
    return phone[0:3] + "#####" + phone[-2:]



from .test_auth_otp import generate_otp


def test_signin_with_multiple_otp_methods(smtpd, testclient, backend, user):
    testclient.app.config["CANAILLE"]["AUTHENTICATION_FACTORS"] = [
        "password",
        "otp",
        "email",
        "sms",
    ]

    res = testclient.get("/login", status=200)

    res.form["login"] = "user"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    res.form["password"] = "correct horse battery staple"
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # TOTP/HOTP
    res.form["otp"] = generate_otp("TOTP", user.secret_token)
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # EMAIL_OTP
    backend.reload(user)
    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302)
    res = res.follow(status=200)

    # SMS_OTP
    backend.reload(user)
    res.form["otp"] = user.one_time_password
    res = res.form.submit(status=302)

    assert (
        "success",
        "Connection successful. Welcome Johnny",
    ) in res.flashes



from flask import Blueprint
from flask import current_app
from flask import flash
from flask import g
from flask import redirect
from flask import request
from flask import url_for

from canaille.app.i18n import gettext as _
from canaille.app.session import logout_user
from canaille.app.templating import render_template
from canaille.core.auth import AuthenticationSession
from canaille.core.auth import get_user_from_login
from canaille.core.auth import login_placeholder
from canaille.core.auth import redirect_to_next_auth_step

from ..forms import LoginForm
from . import email
from . import otp
from . import password
from . import sms

bp = Blueprint("auth", __name__)
bp.register_blueprint(password.bp)
bp.register_blueprint(email.bp)
bp.register_blueprint(sms.bp)
bp.register_blueprint(otp.bp)


@bp.context_processor
def global_processor():
    return {
        "menu": False,
    }


@bp.before_request
def load_auth():
    g.auth = AuthenticationSession.load()


@bp.after_request
def save_auth(response):
    if "auth" in g and g.auth:
        g.auth.save()
    return response


@bp.route("/login", methods=("GET", "POST"))
def login():
    if g.session:
        return redirect(
            url_for("core.account.profile_edition", edited_user=g.session.user)
        )

    form = LoginForm(request.form or None)
    form.render_field_macro_file = "core/partial/login_field.html"
    form["login"].render_kw["placeholder"] = login_placeholder()

    if not request.form or form.form_control():
        return render_template("core/login.html", form=form)

    user = get_user_from_login(form.login.data)
    if user and not user.has_password() and current_app.features.has_smtp:
        return redirect(url_for("core.auth.password.firstlogin", user=user))

    if not form.validate():
        logout_user()
        flash(_("Login failed, please check your information"), "error")
        return render_template("core/login.html", form=form)

    g.auth = AuthenticationSession(user_name=form.login.data)
    return redirect_to_next_auth_step()


@bp.route("/logout")
def logout():
    if user := g.session and g.session.user:
        current_app.logger.security(f"Logout {user.identifier}")

        flash(
            _(
                "You have been disconnected. See you next time %(user)s",
                user=user.name,
            ),
            "success",
        )
        logout_user()
    return redirect("/")
